#include "system.h"
#include "dbint.h"
#include "fcb.h"

/* this file contains the bottom level routines for handling I/O of
   disk blocks (of size BLKSIZE). An LRU cache is used. */

/* HASHSIZE must be power of 2 */

#ifdef CPM
#define BLOCKCOUNT	4	/* number of buffer blocks to get */
#define HASHSIZE	8	/* size of block hash table */
#endif

#ifdef XENIX
#define BLOCKCOUNT	50
#define HASHSIZE	128
#endif

#ifdef MSDOS
#define BLOCKCOUNT	20
#define HASHSIZE	32
#endif

#undef  STATS			/* define for statistics gathering */

extern bool d_open();
extern d_close();
extern bool d_create();
extern bool d_rread();
extern bool d_rwrite();

extern d_fcbName();
extern d_setErr();
extern d_abort();
extern d_2abort();
extern byte *d_alloc();
extern d_free();
extern d_cpPtr();

extern d_rxAlloc();

/* the structure of an I/O descriptor for disk I/O: */

typedef struct io {
    struct io *io_next;
    struct FCB io_fcb;
    bool io_open;
} IO;

/* the structure of a relation I/O descriptor: */

typedef struct riol {
    struct riol *rl_next;	/* next relation in chain */
    RELATION *rl_rel;		/* pointer to RELATION for this one */
    IO *rl_data;		/* descriptors for data files */
    IO *rl_index;		/* descriptors for index files */
    char *rl_ddrv;		/* string of data drive codes */
    char *rl_xdrv;		/* string of index drive codes */
    byte rl_dcnt;		/* number of data drives */
    byte rl_xcnt;		/* number of index drives */
    byte rl_dexis;		/* number of existing data files */
    byte rl_xexis;		/* number of existing index files */
    bool rl_act;		/* set to false when relation closed */
} RIOL;

/* the structure of a block descriptor: */

typedef struct block {
    struct block *b_next;	/* pointer to next block in LRU list */
    RIOL *b_riol;		/* pointer to relation block is in */
    POINTER b_ptr;		/* position in relation of this block */
    bool b_data;		/* true if data block, not index/inode */
    bool b_full;		/* true if block contains something */
    bool b_dirty;		/* true if contents don't match disk */
    byte b_junk;		/* Make sure next byte is 2 byte aligned */
    byte b_buff[BLKSIZE];	/* the actual data for the block */
} BLOCK;

/* the structure of a hash table entry: */

typedef struct {
    RIOL *h_riol;		/* identify the relation */
    BLOCK *h_blk;		/* pointer to block containing it */
    POINTER h_ptr;		/* the pointer hashed here */
    bool h_isData;		/* data/index indicator */
} HASH;

BLOCK *d_Blks;

RIOL *d_RIOList;			/* list of active relations */

HASH HashTable[HASHSIZE];			/* hash table of blocks */

byte *d_BXNext, *d_BXEnd;
POINTER d_BXPtr;
RIOL *d_BXRiol;
bool d_BXEof;

bool d_NewBlock;

RELATION *d_BRelA, *d_BRelB;
long d_BMnPA, d_BMnPB, d_BMxPA, d_BMxPB;
BLOCK *d_BlkA, *d_BlkB;
bool d_BALast;

#ifdef STATS
long d_BBCnt, d_RCnt, d_WCnt, d_GCnt, d_PCnt;
#endif

/*
 * d_bInit - initialize all buffers to the unused state.
 *
 *	NOTE: because of the way that B-tree deletion (file INDEX3.C) works,
 *	      there are situations in which four (4) blocks will be active at
 *	      one time. Thus, the number of buffer blocks MUST be at least
 *	      that number for correct operation.
 */

d_bInit()
{
    BLOCK *b;
    word n;

#ifdef MSDOS
    word count;
    extern long sizmem();
    extern char *getml();

    /* Allocate half of available memory to use as buffers */
    
    allmem();
    count = (sizmem() / 2) / sizeof(BLOCK);
    count = count - count % 5;	/* Round down to multiple of 5 */
    if (count > 50) {
	count = 50;		/* Max of 50 buffers (seems reasonable!) */
    }
    b = nil;
    while (count >= 5) {
	b = (BLOCK *)getml(count * (long)sizeof(BLOCK));
	if (b == nil) {
	    count = count - 5;
	} else {
	    break;
	}
    }

    if (count < 5) {
	d_2abort(19, 0, count);
    }

    d_Blks = nil;
    for (n = 0; n != count; n++) {
	b->b_next = d_Blks;
	b->b_full = false;
	d_Blks = b++;
    }
#endif
#ifdef XENIX
    d_Blks = nil;
    for (n = 0; n != BLOCKCOUNT; n++) {
	b = (BLOCK *)d_alloc(sizeof(BLOCK));
	b->b_next = d_Blks;
	b->b_full = false;
	d_Blks = b;
    }
#endif
#ifdef CPM
    d_Blks = nil;
    for (n = 0; n != BLOCKCOUNT; n++) {
	b = (BLOCK *)d_alloc(sizeof(BLOCK));
	b->b_next = d_Blks;
	b->b_full = false;
	d_Blks = b;
    }
#endif
    d_RIOList = nil;
    d_BRelA = nil;
    d_BRelB = nil;
    d_BALast = true;
    for (n = 0; n != HASHSIZE; n++) {
	HashTable[n].h_riol = nil;
    }
#ifdef STATS
    d_BBCnt = 0l;
    d_RCnt = 0l;
    d_WCnt = 0l;
    d_GCnt = 0l;
    d_PCnt = 0l;
#endif
}

#ifdef MSDOS

/*
 * d_blkCnt - count the number of blocks in use.
 */

int d_blkCnt()
{
    BLOCK *b;
    int i;

    b = d_Blks;
    i = 0;
    while (b != nil) {
	b = b->b_next;
	i++;
    }
    return(i);
}

#endif

#ifdef STATS
/*
 * d_bStats - return the 5 statistics counters.
 */

d_bStats(bb, r, w, g, p)
    long *bb, *r, *w, *g, *p;
{

    *bb = d_BBCnt;
    *r = d_RCnt;
    *w = d_WCnt;
    *g = d_GCnt;
    *p = d_PCnt;
}
#endif

/*
 * d_bTerm - free up all allocated space.
 */

#ifdef CLEANUP
d_bTerm()
{
    RIOL *riol;
    BLOCK *b;

    while (d_RIOList != nil) {
	riol = d_RIOList;
	d_RIOList = d_RIOList->rl_next;
	d_free((byte *)riol, sizeof(RIOL));
    }
    while (d_Blks != nil) {
	b = d_Blks;
	d_Blks = d_Blks->b_next;
	d_free((byte *)b, sizeof(BLOCK));
    }
}
#endif

/*
 * d_bgFile - get the IO descriptor associated with the given file.
 *	      NOTE: fileNum is 1 origin.
 */

IO *
d_bgFile(riol, isData, fileNum)
    RIOL *riol;
    bool isData;
    byte fileNum;
{
    IO *io;

    io = (isData ? riol->rl_data : riol->rl_index);
    while (fileNum-- != 1) {
	io = io->io_next;
    }
    return(io);
}

/*
 * d_bFlush - flush a dirty block.
 */

d_bFlush(b)
    BLOCK *b;
{
    struct FCB *f;
    word blk;

    if (b->b_dirty) {
	b->b_dirty = false;
	f = &d_bgFile(b->b_riol, b->b_data, b->b_ptr.p_file)->io_fcb;
	blk = b->b_ptr.p_block;
#ifdef STATS
	d_WCnt++;
#endif
	if (!d_rwrite(f, blk, &b->b_buff[0], BLKSIZE / SYSSIZE)) {
	    d_2abort(11, f->fcbext[2] - '0', blk);
	}
    }
}

/*
 * d_bfNum - set up a file name with a file number.
 */

d_bfNum(f, drive, name, file)
    struct FCB *f;
    char drive;
    char *name;
    word file;
{
    char numBuff[4];

    numBuff[0] = file / 100 + '0';
    numBuff[1] = (file / 10) % 10 + '0';
    numBuff[2] = file % 10 + '0';
    numBuff[3] = 0;
    d_fcbName(f, drive, name, &numBuff[0]);
}

/*
 * d_bfOpen - local routine to open a data or index file.
 */

d_bfOpen(f, create)
    struct FCB *f;
    bool create;
{

    if (!(create ? d_create(f) : d_open(f))) {
	d_abort(12);
    }
}

/*
 * d_bNewIO - add a new IO node to a given list.
 */

IO *
d_bNewIO(p, drive, name, fyle)
    IO **p;
    char drive;
    char *name;
    word fyle;
{
    IO *io;

    io = (IO *)d_alloc(sizeof(IO));
    io->io_open = true;
    d_bfNum(&io->io_fcb, drive, name, fyle);
    d_bfOpen(&io->io_fcb, true);
    io->io_next = nil;
    while (*p != nil) {
	p = &(*p)->io_next;
    }
    *p = io;
    return(io);
}

/*
 * d_bsNew - set d_NewBlock to indicate that an allocation has been made.
 */

d_bsNew()
{
    d_NewBlock = true;
}

/*
 * d_bcNew - clear d_NewBlock after accessing it.
 */

d_bcNew()
{
    d_NewBlock = false;
}

/*
 * d_brdBlk - read a block from a file.
 */

d_brdBlk(buff, f, blk)
    char *buff;
    struct FCB *f;
    word blk;
{

#ifdef STATS
    d_RCnt++;
#endif
    if (!d_rread(f, blk, buff, BLKSIZE / SYSSIZE)) {
	if (d_NewBlock) {
	    if (!d_rwrite(f, blk, buff, BLKSIZE / SYSSIZE)) {
		d_2abort(14, f->fcbext[2] - '0', blk);
	    }
	} else {
	    d_2abort(15, f->fcbext[2] - '0', blk);
	}
    }
}

/*
 * d_bbGet - get a buffer containing the required block.
 */

BLOCK *
d_bbGet(riol, isData, ptr)
    RIOL *riol;
    bool isData;
    POINTER *ptr;
{
    BLOCK *b, *prev, *pprev;
    IO *io;
    HASH *h1, *h2;
    word block;
    byte fyle;
    bool got;

#ifdef STATS
    d_BBCnt++;
#endif
    fyle = ptr->p_file;
    block = ptr->p_block;

    /* check block request for validity */

    if (isData) {
	if (fyle > riol->rl_dexis && block != 0 ||
		fyle == 0 || fyle > riol->rl_dexis + 1) {
	    d_2abort(23, fyle, block);
	}
    } else {
	if (fyle > riol->rl_xexis && block != 0 ||
		fyle == 0 || fyle > riol->rl_xexis + 1) {
	    d_2abort(24, fyle, block);
	}
    }

    /* quick check for block in hash table */

    h1 = &HashTable[(block + fyle) & (HASHSIZE - 1)];
    if (h1->h_riol == riol && h1->h_isData == isData &&
	    h1->h_ptr.p_file == fyle && h1->h_ptr.p_block == block) {
	return(h1->h_blk);
    }

    /* search for block in active block list */

    b = d_Blks;
    got = false;
    while (!got && b != nil) {
	if (!b->b_full ||
		b->b_riol == riol &&
		b->b_data == isData && b->b_ptr.p_file == fyle &&
		b->b_ptr.p_block == block) {
	    got = true;
	} else {
	    pprev = prev;
	    prev = b;
	    b = b->b_next;
	}
    }
    if (!got) {
	b = prev;
	prev = pprev;
	if (b->b_full) {
	    /* The block buffer we have chosen to replace is occupied. Flush
	       it out if it's dirty, and make sure it isn't referenced by
	       either of the two look-aside mechanisms. */
	    b->b_full = false;
	    d_bFlush(b);
	    if (d_BlkA == b) {
		d_BRelA = nil;
	    }
	    if (d_BlkB == b) {
		d_BRelB = nil;
	    }
	    h2 = &HashTable[(b->b_ptr.p_block + b->b_ptr.p_file) &
				(HASHSIZE - 1)];
	    if (h2->h_blk == b) {
		h2->h_riol = nil;
	    }
	}
    }
    if (!b->b_full) {
	/* The block we are using doesn't currently contain the actual file
	   block we want. Read it in. */
	b->b_full = true;
	b->b_riol = riol;
	b->b_data = isData;
	b->b_ptr.p_file = fyle;
	b->b_ptr.p_block = block;
	if (isData) {
	    if (fyle > riol->rl_dexis) {
		io = d_bNewIO(&riol->rl_data,
			      riol->rl_ddrv[(fyle - 1) % riol->rl_dcnt],
			      riol->rl_rel->r_name, 1000 - (word)fyle);
		riol->rl_dexis = riol->rl_dexis + 1;
	    } else {
		io = d_bgFile(riol, true, fyle);
		if (!io->io_open) {
		    d_bfNum(&io->io_fcb,
			    riol->rl_ddrv[(fyle - 1) % riol->rl_dcnt],
			    riol->rl_rel->r_name, 1000 - (word)fyle);
		    d_bfOpen(&io->io_fcb, false);
		    io->io_open = true;
		}
	    }
	} else {
	    if (fyle > riol->rl_xexis) {
		io = d_bNewIO(&riol->rl_index,
			      riol->rl_xdrv[(fyle - 1) % riol->rl_xcnt],
			      riol->rl_rel->r_name, (word)fyle);
		riol->rl_xexis = riol->rl_xexis + 1;
	    } else {
		io = d_bgFile(riol, false, fyle);
		if (!io->io_open) {
		    d_bfNum(&io->io_fcb,
			    riol->rl_xdrv[(fyle - 1) % riol->rl_xcnt],
			    riol->rl_rel->r_name, (word)fyle);
		    d_bfOpen(&io->io_fcb, false);
		    io->io_open = true;
		}
	    }
	}
	d_brdBlk(&b->b_buff[0], &io->io_fcb, block);
    }
    /* Move this block to the front of the buffer list */
    if (b != d_Blks) {
	prev->b_next = b->b_next;
	b->b_next = d_Blks;
	d_Blks = b;
    }
    /* Enter this block into the hash table */
    h1->h_blk = b;
    h1->h_riol = riol;
    h1->h_ptr.p_file = fyle;
    h1->h_ptr.p_block = block;
    h1->h_isData = isData;
    return(b);
}

/*
 * d_bxhGet - get a byte from the index file header portion.
 */

byte
d_bxhGet()
{

    if (d_BXNext == d_BXEnd) {
	d_BXNext = &d_bbGet(d_BXRiol, false, &d_BXPtr)->b_buff[0];
	d_BXEnd = d_BXNext + (BLKSIZE - sizeof(POINTER));
	d_cpPtr(&d_BXPtr, d_BXEnd);
    }
    return(*d_BXNext++);
}

/*
 * d_bgDrv - get a vector of drive codes.
 */

char *
d_bgDrv(cnt)
    byte cnt;
{
    char *p, *q;

    p = (char *)d_alloc(cnt);
    q = p;
    while (cnt-- != 0) {
	*q++ = d_bxhGet();
    }
    return(p);
}

/*
 * d_bgIO - create a list of IO's.
 */

IO *
d_bgIO(cnt)
    byte cnt;
{
    IO **p;
    IO *io, *first;

    p = &first;
    while (cnt-- != 0) {
	io = (IO *)d_alloc(sizeof(IO));
	io->io_open = false;
	*p = io;
	p = &io->io_next;
    }
    *p = nil;
    return(first);
}

/*
 * d_bOpen - open a relation as far as buffering is concerned.
 */

RIOL *
d_bOpen(drive, r)
    char drive;
    RELATION *r;
{
    RIOL *riol, *frl, *trl;
    IO *io;
    byte cnt;

    frl = nil;
    trl = nil;
    riol = d_RIOList;
    while (riol != nil) {
	if (riol->rl_rel == r && riol->rl_act) {
	    trl = riol;
	} else if (!riol->rl_act) {
	    frl = riol;
	}
	riol = riol->rl_next;
    }
    if (trl != nil) {
	d_abort(16);
    } else {
	io = (IO *)d_alloc(sizeof(IO));
	d_fcbName(&io->io_fcb, drive, r->r_name, "001");
	if (!d_open(&io->io_fcb)) {
	    d_free((byte *)io, sizeof(IO));
	    d_setErr(32);
	    return(nil);
	} else {
	    if (frl == nil) {
		riol = (RIOL *)d_alloc(sizeof(RIOL));
		riol->rl_next = d_RIOList;
		d_RIOList = riol;
	    } else {
		riol = frl;
	    }
	    riol->rl_rel = r;
	    riol->rl_act = true;
	    riol->rl_xcnt = 1;		/* at least ONE drive */
	    riol->rl_xexis = 1;
	    riol->rl_xdrv = &drive;
	    io->io_next = nil;
	    io->io_open = true;
	    riol->rl_index = io;
	    d_BXNext = nil;
	    d_BXEnd = nil;
	    d_BXPtr.p_file = 1;
	    d_BXPtr.p_block = 0;
	    d_BXRiol = riol;
	    if (d_bxhGet() != VERSION) {
		d_abort(7);
	    }
	    cnt = d_bxhGet();		/* validity flag */
	    cnt = d_bxhGet();		/* number of index drives */
	    riol->rl_xcnt = cnt;
	    riol->rl_xdrv = d_bgDrv(cnt);
	    cnt = d_bxhGet();		/* number of index files */
	    riol->rl_xexis = cnt;
	    io->io_next = d_bgIO(cnt - 1);
	    cnt = d_bxhGet();		/* number of data drives */
	    riol->rl_dcnt = cnt;
	    riol->rl_ddrv = d_bgDrv(cnt);
	    cnt = d_bxhGet();		/* number of data files */
	    riol->rl_dexis = cnt;
	    riol->rl_data = d_bgIO(cnt);
	    return(riol);
	}
    }
}

/*
 * d_rbCre - set up a new RIOL block, but for a relation being created.
 */

RIOL *
d_rbCre(r, xdrives, ddrives)
    RELATION *r;
    char *xdrives, *ddrives;
{
    RIOL *riol;
    IO *io;
    POINTER ptr;
    word cnt;

    io = (IO *)d_alloc(sizeof(IO));
    d_fcbName(&io->io_fcb, *xdrives, r->r_name, "001");
    if (d_open(&io->io_fcb)) {
	d_free((byte *)io, sizeof(IO));
	d_setErr(18);
	return(nil);
    } else if (!d_create(&io->io_fcb)) {
	d_free((byte *)io, sizeof(IO));
	d_setErr(19);
	return(nil);
    } else {
	riol = d_RIOList;
	while (riol != nil && riol->rl_act) {
	    riol = riol->rl_next;
	}
	if (riol == nil) {
	    riol = (RIOL *)d_alloc(sizeof(RIOL));
	    riol->rl_next = d_RIOList;
	    d_RIOList = riol;
	}
	riol->rl_rel = r;
	riol->rl_act = true;
	io->io_open = true;
	io->io_next = nil;
	riol->rl_index = io;
	cnt = strlen(xdrives);
	riol->rl_xdrv = (char *)d_alloc(cnt);
	strcpy(riol->rl_xdrv, xdrives);
	riol->rl_xcnt = cnt;
	riol->rl_xexis = 1;
	riol->rl_data = nil;
	cnt = strlen(ddrives);
	riol->rl_ddrv = (char *)d_alloc(cnt);
	strcpy(riol->rl_ddrv, ddrives);
	riol->rl_dcnt = cnt;
	riol->rl_dexis = 0;
	ptr.p_file = 1;
	ptr.p_block = 0;
	d_bsNew();
	((POINTER *)(&d_bbGet(riol, false, &ptr)->b_buff[0] +
	    (BLKSIZE - sizeof(POINTER))))->p_file = 0;
	d_bcNew();
	return(riol);
    }
}

/*
 * d_bxhPut - put a byte to the index file header portion.
 */

d_bxhPut(byt)
    byte byt;
{
    BLOCK *b;

    if (d_BXNext == d_BXEnd) {
	if (d_BXEof) {
	    d_rxAlloc(d_BXRiol->rl_rel, &d_BXPtr);
	    d_cpPtr(d_BXEnd, &d_BXPtr);
	}
	b = d_bbGet(d_BXRiol, false, &d_BXPtr);
	b->b_dirty = true;
	d_BXNext = &b->b_buff[0];
	d_BXEnd = d_BXNext + (BLKSIZE - sizeof(POINTER));
	if (d_BXEof) {
	    ((POINTER *)d_BXEnd)->p_file = 0;
	} else {
	    d_BXEof = (((POINTER *)d_BXEnd)->p_file == 0);
	}
	d_cpPtr(&d_BXPtr, d_BXEnd);
    }
    *d_BXNext++ = byt;
}

/*
 * d_bpDrv - write out (to header) a list of drive codes.
 */

d_bpDrv(p, cnt)
    char *p;
    byte cnt;
{
    d_bxhPut(cnt);
    while (cnt-- != 0) {
	d_bxhPut(*p++);
    }
}

/*
 * d_bCls1 - close a relation as far as buffering is concerned.
 */

d_bCls1(r)
    RELATION *r;
{
    d_BXNext = nil;
    d_BXEnd = nil;
    d_BXPtr.p_file = 1;
    d_BXPtr.p_block = 0;
    d_BXRiol = (RIOL *)r->r_riol;
    d_BXEof = false;
    d_bxhPut(VERSION);		/* version number */
    d_bxhPut(0);		/* validity flag */
    d_bpDrv(d_BXRiol->rl_xdrv, d_BXRiol->rl_xcnt);
    d_bxhPut(d_BXRiol->rl_xexis);
    d_bpDrv(d_BXRiol->rl_ddrv, d_BXRiol->rl_dcnt);
    d_bxhPut(d_BXRiol->rl_dexis);
}

/*
 * d_bclIO - close and free a list of IO's.
 */

d_bclIO(io)
    IO *io;
{
    IO *temp;

    while (io != nil) {
	if (io->io_open) {
	    d_close(&io->io_fcb);
	}
	temp = io;
	io = io->io_next;
	d_free(temp, sizeof(IO));
    }
}

/*
 * d_bCls2 - second half of closing a relation.
 */

d_bCls2(r)
    RELATION *r;
{
    BLOCK **bb;
    BLOCK *b, *freed, *temp;
    word i;

    if (d_BRelA == r) {
	d_BRelA = nil;
    }
    if (d_BRelB == r) {
	d_BRelB = nil;
    }
    d_BXRiol = (RIOL *)r->r_riol;
    for (i = 0; i != HASHSIZE; i++) {
	if (HashTable[i].h_riol == d_BXRiol) {
	    HashTable[i].h_riol = nil;
	}
    }
    freed = nil;
    bb = &d_Blks;
    b = d_Blks;
    while (b != nil) {
	if (b->b_full && b->b_riol == d_BXRiol) {
	    d_bFlush(b);
	    b->b_full = false;
	    temp = b;
	    b = b->b_next;
	    *bb = b;
	    temp->b_next = freed;
	    freed = temp;
	} else {
	    bb = &b->b_next;
	    b = b->b_next;
	}
    }
    *bb = freed;
    d_bclIO(d_BXRiol->rl_index);
    d_bclIO(d_BXRiol->rl_data);
    d_free((byte *)d_BXRiol->rl_xdrv, d_BXRiol->rl_xcnt);
    d_free((byte *)d_BXRiol->rl_ddrv, d_BXRiol->rl_dcnt);
    d_BXRiol->rl_act = false;
}

/*
 * d_bxGet - get a pointer to an index data block.
 */

byte *
d_bxGet(r, ptr)
    RELATION *r;
    POINTER *ptr;
{
    POINTER p;

    p.p_file = ptr->p_file;
    p.p_block = ptr->p_block & (-(BLKSIZE / INOSIZE));
    return(&d_bbGet(r->r_riol, false, &p)->b_buff[
	    (ptr->p_block & (BLKSIZE / INOSIZE - 1)) * INOSIZE]);
}

/*
 * d_bxMark - mark an index data block (or inode block in it) as dirty.
 */

d_bxMark(b, ptr)
    byte *b;
    POINTER *ptr;
{
    byte *p1, *p2;

    b = b - (ptr->p_block & (BLKSIZE / INOSIZE - 1)) * INOSIZE;
    p1 = (byte *)&d_Blks->b_buff[0];
    p2 = (byte *)d_Blks;
    ((BLOCK *)(b - (p1 - p2)))->b_dirty = true;
}

/*
 * d_bMap - map a byte address into a data POINTER and a block offset.
 */

word
d_bMap(pos, ptr, r)
    long pos;
    POINTER *ptr;
    RELATION *r;
{
    long fpos, perFile;

    perFile = (long)r->r_fSize * 1024l;
    ptr->p_file = pos / perFile + 1;
    fpos = pos % perFile;
    ptr->p_block = fpos / (long)BLKSIZE * ((long)BLKSIZE / 128l);
    return((word)(fpos % (long)BLKSIZE));
}

/*
 * d_bdGet - get a byte from a data record.
 */

byte
d_bdGet(r, pos)
    RELATION *r;
    long pos;
{
    BLOCK *b;
    word offs;
    POINTER ptr;

    if (r == d_BRelA && pos <= d_BMxPA && pos >= d_BMnPA) {
	d_BALast = true;
	return(d_BlkA->b_buff[pos - d_BMnPA]);
    } else if (r == d_BRelB && pos <= d_BMxPB && pos >= d_BMnPB) {
	d_BALast = false;
	return(d_BlkB->b_buff[pos - d_BMnPB]);
    } else {
	offs = d_bMap(pos, &ptr, r);
	b = d_bbGet(r->r_riol, true, &ptr);
	if (d_BALast) {
	    d_BRelB = r;
	    d_BMnPB = pos - offs;
	    d_BMxPB = d_BMnPB + (BLKSIZE - 1);
	    d_BlkB = b;
	} else {
	    d_BRelA = r;
	    d_BMnPA = pos - offs;
	    d_BMxPA = d_BMnPA + (BLKSIZE - 1);
	    d_BlkA = b;
	}
	d_BALast = !d_BALast;
	return(b->b_buff[offs]);
    }
}

/*
 * d_bdPut - put a byte to a data record.
 */

d_bdPut(r, pos, byt)
    RELATION *r;
    long pos;
    byte byt;
{
    BLOCK *b;
    word offs;
    POINTER ptr;

    if (r == d_BRelA && pos <= d_BMxPA && pos >= d_BMnPA) {
	d_BALast = true;
	d_BlkA->b_dirty = true;
	d_BlkA->b_buff[pos - d_BMnPA] = byt;
    } else if (r == d_BRelB && pos <= d_BMxPB && pos >= d_BMnPB) {
	d_BALast = false;
	d_BlkB->b_dirty = true;
	d_BlkB->b_buff[pos - d_BMnPB] = byt;
    } else {
	offs = d_bMap(pos, &ptr, r);
	b = d_bbGet(r->r_riol, true, &ptr);
	b->b_buff[offs] = byt;
	b->b_dirty = true;
	if (d_BALast) {
	    d_BRelB = r;
	    d_BMnPB = pos - offs;
	    d_BMxPB = d_BMnPB + (BLKSIZE - 1);
	    d_BlkB = b;
	} else {
	    d_BRelA = r;
	    d_BMnPA = pos - offs;
	    d_BMxPA = d_BMnPA + (BLKSIZE - 1);
	    d_BlkA = b;
	}
	d_BALast = !d_BALast;
    }
}

/*
 * d_bgStuff - get some stuff from a relation.
 */

d_bgStuff(r, pos, ptr, len)
    RELATION *r;
    long pos;
    byte *ptr;
    word len;
{
    byte *p;
    word left;

#ifdef STATS
    d_GCnt++;
#endif
    while (len-- != 0) {
	*ptr++ = d_bdGet(r, pos++);
	if (d_BALast) {
	    left = d_BMxPA + 1l - pos;
	    p = &d_BlkA->b_buff[pos - d_BMnPA];
	    pos = d_BMxPA + 1l;
	} else {
	    left = d_BMxPB + 1l - pos;
	    p = &d_BlkB->b_buff[pos - d_BMnPB];
	    pos = d_BMxPB + 1l;
	}
	if (len < left) {
	    left = len;
	}
	COPY(ptr, p, left);
	ptr += left;
	p += left;
	len -= left;
	if (len != 0) {
	    d_BALast = !d_BALast;
	}
    }
}

/*
 * d_bgStr - get a string from a relation.
 */

d_bgStr(r, pos, ptr, len)
    RELATION *r;
    long pos;
    byte *ptr;
    word len;
{
    byte *p;
    word left;
    byte b;

#ifdef STATS
    d_GCnt++;
#endif
    while (len-- != 0) {
	b = d_bdGet(r, pos++);
	*ptr++ = b;
	if (b == 0) {
	    len = 0;
	} else {
	    if (d_BALast) {
		left = d_BMxPA + 1l - pos;
		p = &d_BlkA->b_buff[pos - d_BMnPA];
		pos = d_BMxPA + 1l;
	    } else {
		left = d_BMxPB + 1l - pos;
		p = &d_BlkB->b_buff[pos - d_BMnPB];
		pos = d_BMxPB + 1l;
	    }
	    while (left-- != 0 && len != 0) {
		len--;
		b = *p++;
		*ptr++ = b;
		if (b == 0) {
		    len = 0;
		}
	    }
	    if (len != 0) {
		d_BALast = !d_BALast;
	    }
	}
    }
}

/*
 * d_bpStuff - put some bytes to a data record.
 */

d_bpStuff(r, pos, ptr, len)
    RELATION *r;
    long pos;
    byte *ptr;
    word len;
{
    byte *p;
    word left;

#ifdef STATS
    d_PCnt++;
#endif
    while (len-- != 0) {
	d_bdPut(r, pos++, *ptr++);
	if (d_BALast) {
	    left = d_BMxPA + 1l - pos;
	    p = &d_BlkA->b_buff[pos - d_BMnPA];
	    pos = d_BMxPA + 1l;
	} else {
	    left = d_BMxPB + 1l - pos;
	    p = &d_BlkB->b_buff[pos - d_BMnPB];
	    pos = d_BMxPB + 1l;
	}
	if (len < left) {
	    left = len;
	}
	COPY(p, ptr, left);
	p += left;
	ptr += left;
	len -= left;
	if (len != 0) {
	    d_BALast = !d_BALast;
	}
    }
}
