#include "system.h"
#include "dbint.h"
#include "scan.h"

/* top level of index handling routines */

extern d_abort();
extern d_error();
extern d_setErr();
extern byte *d_alloc();
extern d_free();

extern RELATION *d_rpget2();

extern INDEX *d_sChk0();
extern INDEX *d_sChk();

extern d_xSetup();
extern long d_xFnd0();
extern SCANLIST *d_xScn1();
extern word d_xScn2();
extern d_xScn3();
extern SCANLIST *d_xScn4();
extern SCANLIST *d_xScn5();

#define SC_ONE	0		/* only one matching element */
#define SC_FALL 1		/* scanning an inode tree for values */
#define SC_SUN	2		/* scan index tree, one value from each key*/
#define SC_SCAN 3		/* scan index tree, return all records */

typedef struct scan {
    struct scan *sc_next;
    RELATION *sc_rel;
    SCANLIST *sc_sl, *sc_slDown;
    long sc_tuple;
    int sc_code;		/* user supplied code for this scan */
    byte sc_mode;
    bool sc_amDown;
} SCAN;

SCAN *d_SCPool;			/* free-list of SCAN's */
SCAN *d_SCAct;			/* list of active SCAN's */
bool d_SCStk;			/* true -> same code scans will stack */

/*
 * d_scInit - initialize for any scanning.
 */

d_scInit()
{
    d_SCAct = nil;
    d_SCPool = nil;
    d_SCStk = false;
}

/*
 * d_scTerm - termination routine for these structures.
 */

#ifdef CLEANUP
d_scTerm()
{
    register SCAN *temp;

    while (d_SCAct != nil) {
	temp = d_SCAct;
	d_SCAct = d_SCAct->sc_next;
	d_free(temp, sizeof(SCAN));
    }
    while (d_SCPool != nil) {
	temp = d_SCPool;
	d_SCPool = d_SCPool->sc_next;
	d_free(temp, sizeof(SCAN));
    }
}
#endif

/*
 * scStack - allow/disallow same code scans to stack.
 */

scStack(canStack)
bool canStack;
{

    d_SCStk = canStack;
}

/*
 * d_scAlloc - allocate a SCAN structure.
 */

SCAN *
d_scAlloc()
{
    register SCAN *sc;

    if (d_SCPool != nil) {
	sc = d_SCPool;
	d_SCPool = d_SCPool->sc_next;
    } else {
	sc = (SCAN *)d_alloc(sizeof(SCAN));
    }
    sc->sc_next = d_SCAct;
    d_SCAct = sc;
    return(sc);
}

/*
 * d_scFree - free a SCAN structure.
 */

d_scFree(sc)
    register SCAN *sc;
{
    sc->sc_next = d_SCPool;
    d_SCPool = sc;
}

/*
 * d_srChk - check to make sure no other scan active for a given code.
 */

d_srChk(code)
int code;
{
    register SCAN *sc;

    if (!d_SCStk) {
	sc = d_SCAct;
	while (sc != nil && sc->sc_code != code) {
	    sc = sc->sc_next;
	}
	if (sc != nil) {
	    d_error(30);
	}
    }
}

/*
 * d_scDTup - when a tuple is being deleted, clear up any SC_ONE scans that
 *	would want to return it.
 */

d_scDTup(tuple)
long tuple;
{
    register SCAN *sc;

    sc = d_SCAct;
    while (sc != nil) {
	if (sc->sc_mode == SC_ONE && sc->sc_tuple == tuple) {
	    sc->sc_tuple = 0L;
	}
	sc = sc->sc_next;
    }
}

/*
 * d_faChk - common routine to check a 'fAll' and start it.
 */

bool
d_faChk(code, relNum, fldNum, key, typ)
    int code;
    word relNum;
    int fldNum;
    byte *key;
    byte typ;
{
    static RELATION *r;
    register INDEX *x;
    static SCAN *sc;
    static long zero, tuple;

    d_srChk(code);
    x = d_sChk(relNum, fldNum, &r, typ);
    d_xSetup(r, x);
    tuple = d_xFnd0(key, &x->x_root);
    zero = 0l;
    if (tuple == zero) {
	return(false);
    } else {
	sc = d_scAlloc();
	sc->sc_rel = r;
	sc->sc_code = code;
	if (tuple > zero) {
	    sc->sc_tuple = tuple;
	    sc->sc_mode = SC_ONE;
	} else {
	    sc->sc_mode = SC_FALL;
	    sc->sc_sl = d_xScn4(r, x, key, true);
	}
	return(true);
    }
}

/*
 * user interface routines for finding all tuples with a given key.
 */

bool
fAllBool(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    bool valu;
{
    return(d_faChk(code, relNum, fldNum, &valu, FTBOOL));
}

bool
fAllShort(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    short valu;
{
    return(d_faChk(code, relNum, fldNum, &valu, FTSHORT));
}

bool
fAllLong(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    long valu;
{
    return(d_faChk(code, relNum, fldNum, &valu, FTLONG));
}

bool
fAllInt(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    int valu;
{
    return(d_faChk(code, relNum, fldNum, &valu, 
		   sizeof(int) == sizeof(long) ? FTLONG : FTSHORT
		  )
    );
}

bool
fAllFloat(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    double valu;
{
    return(d_faChk(code, relNum, fldNum, &valu, FTFLOAT));
}

bool
fAllChar(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    byte *valu;
{
    return(d_faChk(code, relNum, fldNum, valu, FTCHAR));
}

bool
fAllString(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    char *valu;
{
    return(d_faChk(code, relNum, fldNum, valu, FTSTRING));
}

/*
 * d_scGo - start a unique key scan of some kind.
 */

bool
d_scGo(code, r, sl)
    int code;
    RELATION *r;
    SCANLIST *sl;
{
    register SCAN *sc;

    if (sl == nil) {
	return(false);
    } else {
	sc = d_scAlloc();
	sc->sc_rel = r;
	sc->sc_sl = sl;
	sc->sc_code = code;
	sc->sc_mode = SC_SUN;
	return(true);
    }
}

/*
 * d_scnNP - common code for full next/prev scan of unique key values.
 */

bool
d_scnNP(code, relNum, fldNum, leftToRight)
    int code;
    word relNum;
    int fldNum;
    bool leftToRight;
{
    static RELATION *r;
    register INDEX *x;

    x = d_sChk0(relNum, fldNum, &r);
    d_srChk(code);
    return(d_scGo(code, r, d_xScn1(r, x, leftToRight)));
}

/*
 * strtHigh - start a scan of unique key values at top end of index.
 */

bool
strtHigh(code, relNum, fldNum)
    int code;
    word relNum;
    int fldNum;
{
    return(d_scnNP(code, relNum, fldNum, false));
}

/*
 * strtLow - start a scan of unique key values at bottom end of index.
 */

bool
strtLow(code, relNum, fldNum)
    int code;
    word relNum;
    int fldNum;
{
    return(d_scnNP(code, relNum, fldNum, true));
}

/*
 * d_uscChk - set up a scan of the unique key values, starting with some key.
 */

bool
d_uscChk(code, relNum, fldNum, key, typ)
    int code;
    word relNum;
    int fldNum;
    byte *key;
    byte typ;
{
    static RELATION *r;
    register INDEX *x;

    x = d_sChk(relNum, fldNum, &r, typ);
    d_srChk(code);
    return(d_scGo(code, r, d_xScn5(r, x, key)));
}

/*
 * user interface routines for scanning an index tree from a start point.
 */

bool
scanBool(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    bool valu;
{
    return(d_uscChk(code, relNum, fldNum, &valu, FTBOOL));
}

bool
scanShort(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    short valu;
{
    return(d_uscChk(code, relNum, fldNum, &valu, FTSHORT));
}

bool
scanLong(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    long valu;
{
    return(d_uscChk(code, relNum, fldNum, &valu, FTLONG));
}

bool
scanInt(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    int valu;
{
    return(d_uscChk(code, relNum, fldNum, &valu,
		    sizeof(int) == sizeof(long) ? FTLONG : FTSHORT
		   )
    );
}

bool
scanFloat(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    double valu;
{
    return(d_uscChk(code, relNum, fldNum, &valu, FTFLOAT));
}

bool
scanChar(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    byte *valu;
{
    return(d_uscChk(code, relNum, fldNum, valu, FTCHAR));
}

bool
scanString(code, relNum, fldNum, valu)
    int code;
    word relNum;
    int fldNum;
    char *valu;
{
    return(d_uscChk(code, relNum, fldNum, valu, FTSTRING));
}

/*
 * d_scFind - find a scan associated with a given code.
 */

SCAN *
d_scFind(code)
    int code;
{
    register SCAN *sc;

    sc = d_SCAct;
    while (sc != nil && sc->sc_code != code) {
	sc = sc->sc_next;
    }
    if (sc == nil) {
	d_error(31);
    }
    return(sc);
}

/*
 * setAll - set the mode of the given scan to indicate that it should
 *	    go down into inodes to return tuple id's.
 */

setAll(code)
int code;
{
    SCAN *sc;

    sc = d_scFind(code);
    if (sc->sc_mode != SC_SUN) {
	d_error(42);
    }
    sc->sc_mode = SC_SCAN;
    sc->sc_amDown = false;
}

/*
 * d_scDel - delete a scan from the list of active scans.
 */

d_scDel(sc)
    SCAN *sc;
{
    register SCAN **p;

    p = &d_SCAct;
    while (*p != nil && *p != sc) {
	p = &(*p)->sc_next;
    }
    if (*p == nil) {
	d_abort(25);
    }
    *p = sc->sc_next;
    d_scFree(sc);
}

/*
 * d_fNP - common code for fNext/fPrev.
 */

long
d_fNP(code, leftToRight)
    int code;
    bool leftToRight;
{
    register SCAN *sc;
    static long temp;
    static long zero;

    sc = d_scFind(code);
    zero = 0L;
    switch (sc->sc_mode) {
    case SC_ONE:
	/* return the single tuple available for an 'fAll' scan: */
	temp = sc->sc_tuple;
	if (temp == zero) {
	    d_scDel(sc);
	} else {
	    sc->sc_tuple = zero;
	}
	return(temp);
    case SC_FALL:
    case SC_SUN:
	if (d_xScn2(sc->sc_sl, nil, &temp, leftToRight) == 0) {
	    d_xScn3(sc->sc_sl);
	    d_scDel(sc);
	    return(zero);
	} else {
	    return(temp);
	}
    case SC_SCAN:
	if (sc->sc_amDown) {
	    if (d_xScn2(sc->sc_slDown, nil, &temp, leftToRight) == 0) {
		d_xScn3(sc->sc_slDown);
		sc->sc_amDown = false;
	    } else {
		return(temp);
	    }
	}
	if (d_xScn2(sc->sc_sl, nil, &temp, leftToRight) == 0) {
	    d_xScn3(sc->sc_sl);
	    d_scDel(sc);
	    return(zero);
	} else {
	    if (temp < zero) {
		sc->sc_amDown = true;
		sc->sc_slDown = d_xScn4(sc->sc_rel, sc->sc_sl->sl_index,
					&sc->sc_sl->sl_key[0], leftToRight);
		if (sc->sc_slDown == nil ||
			d_xScn2(sc->sc_slDown,nil,&temp,leftToRight) == 0) {
		    d_abort(27);
		}
	    }
	    return(temp);
	}
    default:
	d_abort(0);
    }
}

/*
 * fNext - return the next tuple in a search.
 */

long
fNext(code)
    int code;
{
    return(d_fNP(code, true));
}

/*
 * fPrev - return the previous tuple in a search.
 */

long
fPrev(code)
    int code;
{
    return(d_fNP(code, false));
}

/*
 * d_scNP - common code for scNext/scPrev.
 */

word
d_scNP(code, pKey, pTuple, leftToRight)
    int code;
    byte *pKey;
    long *pTuple;
    bool leftToRight;
{
    register SCAN *sc;
    static word depth;

    sc = d_scFind(code);
    if (sc->sc_mode != SC_SUN) {
	d_error(34);
    }
    depth = d_xScn2(sc->sc_sl, pKey, pTuple, leftToRight);
    if (depth == 0) {
	d_xScn3(sc->sc_sl);
	d_scDel(sc);
	return(0);
    } else {
	return(depth);
    }
}

/*
 * scNext - return the next key-value pair in an index scan.
 */

word
scNext(code, pKey, pTuple)
    int code;
    byte *pKey;
    long *pTuple;
{
    return(d_scNP(code, pKey, pTuple, true));
}

/*
 * scPrev - return the previous key-value pair in an index scan.
 */

word
scPrev(code, pKey, pTuple)
    int code;
    byte *pKey;
    long *pTuple;
{
    return(d_scNP(code, pKey, pTuple, false));
}

/*
 * fAbort - abort a scan.
 */

fAbort(code)
    int code;
{
    register SCAN *sc;

    sc = d_scFind(code);
    switch (sc->sc_mode) {
    case SC_FALL:
    case SC_SUN:
	d_xScn3(sc->sc_sl);
	break;
    case SC_SCAN:
	if (sc->sc_amDown) {
	    d_xScn3(sc->sc_slDown);
	}
	d_xScn3(sc->sc_sl);
	break;
    }
    d_scDel(sc);
}
